@@ -34,17 +34,25 @@ def get_resource_model(context):
     return resource
 
 def get_resource(context):
+    def get_resource_id(context):
+        if hasattr(context, "resource_id"):
+            return context.resource_id
+        else:
+            return get_resource_id(context.parent)
+
     model = get_resource_model(context)
-    if hasattr(context, "resource_id"):
-        resource_id = context.resource_id
-    else:
-        resource_id = context.parent.resource_id
+    resource_id = get_resource_id(context)
     query_filter = methodcaller("filter", model.id == resource_id)
     return get_data(model, query_filter)
 
+def get_context_or_parent(context, context_type):
+    if context.type == context_type:
+        return context
+    else:
+        return get_context_or_parent(context.parent, context_type)
 
 def get_context_data(context, context_type, keys):
-    if context.type != context_type:
-        return get_context_data(context.parent, context_type, keys)
-    else:
+    if context.type == context_type:
         return attrgetter(*keys)(context)
+    else:
+        return get_context_data(context.parent, context_type, keys)
