@@ -7,12 +7,15 @@ from operator import methodcaller
 import colander
 import validictory
 from deform.widget import HiddenWidget
+from deform.widget import PasswordWidget
 from deform.widget import TextAreaWidget
 from deform.widget import TextInputWidget
 from deform.widget import SelectWidget
 from kotti.views.edit import ContentSchema
 from kotti.views.edit import generic_edit
 from kotti.views.edit import generic_add
+from kotti.views.form import CommaSeparatedListWidget
+from kotti.views.form import ObjectType
 from kotti.views.util import ensure_view_selector
 from kotti.views.util import template_api
 from pyramid.response import Response
@@ -39,6 +42,7 @@ from kotti_mapreduce.util.helper import get_object_attributes
 from kotti_mapreduce.util.helper import get_temporary_file
 from kotti_mapreduce.util.model import get_all_data
 from kotti_mapreduce.util.model import get_context_data
+from kotti_mapreduce.util.model import get_context_or_parent
 from kotti_mapreduce.util.model import get_resource
 from kotti_mapreduce.util.model import get_resource_model
 
@@ -88,6 +92,17 @@ _SUPPORT_ACTION_ON_FAILURE = [
     u'TERMINATE_JOB_FLOW',
 ]
 
+_SUPPORT_INSTANCE_TYPE = [
+    (u'm1.small', u'Small (m1.small)'),
+    (u'c1.medium', u'High-CPU Medium (c1.medium)'),
+    (u'm1.large', u'Large (m1.large)'),
+    (u'm1.xlarge', u'Extra Large (m1.xlarge)'),
+    (u'c1.xlarge', u'High-CPU Extra Large (c1.xlarge)'),
+    (u'm2.xlarge', u'High-Memory Extra Large (m2.xlarge)'),
+    (u'm2.2xlarge', u'High-Memory Double Extra Large (m2.2xlarge)'),
+    (u'm2.4xlarge', u'High-Memory Quad Extra Large (m2.4xlarge)'),
+]
+
 
 class EMRJobResourceSchema(ContentSchema):
     aws_access_key_id = colander.SchemaNode(
@@ -99,6 +114,7 @@ class EMRJobResourceSchema(ContentSchema):
         colander.String(),
         title=_(u'AWS secret access key'),
         description=_(u'AWS secret access key to use AWS API.'),
+        widget=PasswordWidget(),
     )
     region = colander.SchemaNode(
         colander.String(),
@@ -109,12 +125,14 @@ class EMRJobResourceSchema(ContentSchema):
         colander.String(),
         title=_(u'Master instance type'),
         description=_(u'EC2 instance type of the master node.'),
+        widget=SelectWidget(values=_SUPPORT_INSTANCE_TYPE),
         default='m1.small',
     )
     slave_instance_type = colander.SchemaNode(
         colander.String(),
         title=_(u'Slave instance type'),
         description=_(u'EC2 instance type of the slave nodes.'),
+        widget=SelectWidget(values=_SUPPORT_INSTANCE_TYPE),
         default='m1.small',
     )
     ec2_keyname = colander.SchemaNode(
@@ -180,7 +198,7 @@ class EMRJobResourceSchema(ContentSchema):
         description=_(u'Version of Hadoop to use. This no longer.'),
         missing=u'',
     )
-    hive_version = colander.SchemaNode(
+    hive_versions = colander.SchemaNode(
         colander.String(),
         title=_(u'Hive version'),
         description=_(u'Version of Hive to use.'),
@@ -198,6 +216,7 @@ def add_emrjob_resource(context, request):
 def view_emrjob_resource(context, request):
     return {
         'api': template_api(context, request),
+        'resource_models': [EMRJobResource, JobService, Bootstrap],
     }
 
 
@@ -208,7 +227,8 @@ def deferred_resource_data(node, kw):
     resource = get_resource_model(context)
     if resource:
         order_by = methodcaller('order_by', asc(resource.title))
-        data = get_all_data(resource, context, order_by=order_by)
+        job_container = get_context_or_parent(context, u'jobcontainer')
+        data = get_all_data(resource, job_container, order_by=order_by)
         values = map(attrgetter('id', 'title'), data)
     return SelectWidget(values=values)
 
@@ -234,27 +254,46 @@ def view_jobservice(context, request):
         'bootstraps': False,
         'is_runnable': False,
         'resource': get_resource(context),
+        'resource_models': [EMRJobResource, JobService, Bootstrap],
     }
 
 
+@colander.deferred
+def deferred_bootstrap_validator(node, kw):
+    def raise_invalid_bootstrap(node, value):
+        raise colander.Invalid(
+            node, _(u"The Bootstrap is not found."))
+
+    request = kw['request']
+    if request.POST:
+        parent = request.context.parent
+        context = get_context_or_parent(parent, u'jobcontainer')
+        bootstraps = get_all_data(Bootstrap, context)
+        availables = [bootstrap.title for bootstrap in bootstraps]
+        posted_bootstraps = request.params.get('bootstrap_titles')
+        if posted_bootstraps:
+            for posted_bootstrap in posted_bootstraps.split(','):
+                if posted_bootstrap not in availables:
+                    return raise_invalid_bootstrap
+
+@colander.deferred
+def deferred_bootstrap_widget(node, kw):
+    parent = kw['request'].context.parent
+    context = get_context_or_parent(parent, u'jobcontainer')
+    bootstraps = get_all_data(Bootstrap, context)
+    availables = [bootstrap.title.encode('utf-8') for bootstrap in bootstraps]
+    widget = CommaSeparatedListWidget(
+        template='kotti_mapreduce:templates/deform/bootstrap-list',
+        available_bootstraps=availables,
+    )
+    return widget
+
 _SUPPORT_JOBTYPE = [
     u'hive',
     u'custom-jar',
     u'streaming',
 ]
 
-@colander.deferred
-def deferred_bootstrap_data(node, kw):
-    context = kw['request'].context
-    rsc_and_bs = []
-    if context.type == u'jobservice':
-        rsc_and_bs = context.parent.children
-    elif context.type == u'jobflow':
-        rsc_and_bs = context.parent.parent.children
-    bootstraps = filter(lambda c: c.type == 'bootstrap', rsc_and_bs)
-    values = [(u'', u'')] + map(attrgetter('id', 'title'), bootstraps)
-    return SelectWidget(values=values)
-
 class JobFlowSchema(ContentSchema):
     jobflow_id = colander.SchemaNode(
         colander.String(),
@@ -280,118 +319,13 @@ class JobFlowSchema(ContentSchema):
         description=_(u'Use metastore located outside of the cluster.'),
         missing=u'',
     )
-    # FIXME: implement one to many relation
-    bootstrap1_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #1'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap2_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #2'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap3_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #3'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap4_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #4'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap5_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #5'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap6_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #6'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap7_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #7'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap8_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #8'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap9_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #9'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap10_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #10'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap11_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #11'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap12_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #12'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap13_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #13'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap14_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #14'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap15_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #15'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
-    )
-    bootstrap16_id = colander.SchemaNode(
-        colander.String(),
-        title=_(u'Bootstrap #16'),
-        widget=deferred_bootstrap_data,
-        default=u'',
-        missing=u'',
+    bootstrap_titles = colander.SchemaNode(
+        ObjectType(),
+        title=_('Bootstraps'),
+        description=_(u'Input a bootstrap name registered in advance.'),
+        validator=deferred_bootstrap_validator,
+        widget=deferred_bootstrap_widget,
+        missing=[],
     )
 
 @ensure_view_selector
@@ -399,7 +333,8 @@ def edit_jobflow(context, request):
     return generic_edit(context, request, JobFlowSchema())
 
 def add_jobflow(context, request):
-    return generic_add(context, request, JobFlowSchema(), JobFlow, u'jobflow')
+    return generic_add(context, request, JobFlowSchema(),
+                       JobFlow, JobFlow.type_info.title)
 
 def view_jobflow_with_ajax(context, request):
     req_type = request.params.get('type')
@@ -433,10 +368,12 @@ def view_jobflow(context, request):
         'bootstraps': get_bootstraps(context),
         'is_runnable': is_runnable_jobflow(context),
         'resource': get_resource(context),
+        'resource_models': [JobFlow],
     }
 
 
 _SUPPORT_ACTION_TYPE = [
+    ('', u''),
     ('hadoop', u'Configure Hadoop'),
     ('daemon', u'Configure Daemons'),
     ('ganglia', u'Install Ganglia'),
@@ -445,17 +382,13 @@ _SUPPORT_ACTION_TYPE = [
     ('custom', u'Custom Action'),
 ]
 
-@colander.deferred
-def deferred_jobflow_id(node, kw):
-    jobflow_id = kw['request'].context.jobflow_id
-    return jobflow_id
-
 class BootstrapSchema(ContentSchema):
     action_type = colander.SchemaNode(
         colander.String(),
         title=_(u'Action Type'),
         description=_(u'The action type.'),
         widget=SelectWidget(values=_SUPPORT_ACTION_TYPE),
+        default=u'',
     )
     path_uri = colander.SchemaNode(
         colander.String(),
@@ -482,6 +415,7 @@ def add_bootstrap(context, request):
 def view_bootstrap(context, request):
     return {
         'api': template_api(context, request),
+        'resource_models': [EMRJobResource, JobService, Bootstrap],
     }
 
 
@@ -528,7 +462,11 @@ def deferred_jobstep_widget(node, kw):
 def deferred_default_step_args(node, kw):
     jobtype = get_context_data(kw['request'].context, 'jobflow', ['jobtype'])
     if jobtype == u'hive':
-        return _DEFAULT_HIVE_STEP_ARGS
+        resource = get_resource(kw['request'].context)
+        args = _DEFAULT_HIVE_STEP_ARGS
+        if resource.hive_versions:
+            args += u'--hive-versions\n{0}\n'.format(resource.hive_versions)
+        return args
     elif jobtype == u'custom-jar':
         return _DEFAULT_CUSTOM_JAR_STEP_ARGS
     else:
@@ -546,6 +484,11 @@ s3://us-east-1.elasticmapreduce/libs/hive/
 _DEFAULT_CUSTOM_JAR_STEP_ARGS = """
 """
 
+@colander.deferred
+def deferred_jobflow_id(node, kw):
+    jobflow_id = kw['request'].context.jobflow_id
+    return jobflow_id
+
 def step_args_validator(node, value):
     if value:
         try:
@@ -622,6 +565,7 @@ def view_jobstep(context, request):
     return {
         'api': template_api(context, request),
         'attr_names': attr_names,
+        'resource_models': [JobStep],
     }
 
 def view_download_file(context, request):
